/*
 *  Copyright (C) 2012, Northwestern University
 *  See COPYRIGHT notice in top-level directory.
 *
 *  $Id: test_bput.c 1628 2014-05-12 20:13:06Z wkliao $
 */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <pnetcdf.h>

#define FILE_NAME "testfile.nc"

#define ERRCODE 2
#define ERR(e) {printf("Error at line %d: err=%d %s\n", __LINE__, e, ncmpi_strerror(e)); exit(ERRCODE);}

/*----< main() >------------------------------------------------------------*/
int main(int argc, char **argv) {
    int i, j, ncid, dimid[2], varid, retval, errs=0, rank, nprocs, verbose;
    int req[2], status[2];
    float  var[4][6];
    char *filename="testfile.nc";
    MPI_Offset bufsize,  start[2], count[2], stride[2], imap[2];
    MPI_Info info;

    MPI_Init(&argc, &argv);
    MPI_Comm_rank(MPI_COMM_WORLD, &rank);
    MPI_Comm_size(MPI_COMM_WORLD, &nprocs);

    verbose = 0;
    if (nprocs > 1 && rank == 0 && verbose)
        printf("Warning: %s is designed to run on 1 process\n", argv[0]);

    if (argc > 2) {
        if (!rank) printf("Usage: %s [filename]\n",argv[0]);
        MPI_Finalize();
        return 0;
    }
    if (argc == 2) filename = argv[1];

    MPI_Info_create(&info);
    /* MPI_Info_set(info, "romio_pvfs2_posix_write","enable"); */

    if (NC_NOERR != (retval = ncmpi_create(MPI_COMM_WORLD, filename,
        NC_CLOBBER | NC_64BIT_DATA, info, &ncid)))
       ERR(retval);
    MPI_Info_free(&info);

    /* define a variable of a 6 x 4 integer array in the nc file */
    if (NC_NOERR != (retval = ncmpi_def_dim(ncid, "Y", 6, &dimid[0]))) ERR(retval);
    if (NC_NOERR != (retval = ncmpi_def_dim(ncid, "X", 4, &dimid[1]))) ERR(retval);
    if (NC_NOERR != (retval = ncmpi_def_var(ncid, "var", NC_INT64, 2, dimid, &varid)))
        ERR(retval);
    if (NC_NOERR != (retval = ncmpi_enddef(ncid))) ERR(retval);

    /* set the contents of the write buffer var, a 4 x 6 float array
          50, 51, 52, 53, 54, 55,
          56, 57, 58, 59, 60, 61,
          62, 63, 64, 65, 66, 67,
          68, 69, 70, 71, 72, 73
     */
    for (j=0; j<4; j++) for (i=0; i<6; i++) var[j][i] = j*6+i + 50;

    /* bufsize must be max of data type converted before and after */
    bufsize = 4*6*sizeof(long long);
    if (NC_NOERR != (retval = ncmpi_buffer_attach(ncid, bufsize)))
        ERR(retval);

    /* write var to the NC variable in the matrix transposed way */
    count[0]  = 6; count[1]  = 2;
    stride[0] = 1; stride[1] = 1;
    imap[0]   = 1; imap[1]   = 6;   /* would be {4, 1} if not transposing */

    if (rank > 0) /* non-root processes just participate the call */
        count[0] = count[1] = 0;

    /* write the first two columns of the NC variable in the matrix transposed way */
    start[0]  = 0; start[1]  = 0;
    if (NC_NOERR != (retval = ncmpi_bput_varm_float(ncid, varid, start, count, stride, imap, &var[0][0], &req[0])))
        ERR(retval);

    /* write the second two columns of the NC variable in the matrix transposed way */
    start[0]  = 0; start[1]  = 2;
    if (NC_NOERR != (retval = ncmpi_bput_varm_float(ncid, varid, start, count, stride, imap, &var[2][0], &req[1])))
        ERR(retval);

    if (NC_NOERR != (retval = ncmpi_wait_all(ncid, 2, req, status)))
        ERR(retval);

    /* check each bput status */
    for (i=0; i<2; i++)
        if (status[i] != NC_NOERR) ERR(status[i]);

    if (NC_NOERR != (retval = ncmpi_buffer_detach(ncid)))
        ERR(retval);

    /* the output from command "ncmpidump -v var test.nc" should be:
           var =
            50, 56, 62, 68,
            51, 57, 63, 69,
            52, 58, 64, 70,
            53, 59, 65, 71,
            54, 60, 66, 72,
            55, 61, 67, 73 ;
     */

    /* check if the contents of write buffer have been altered (should not be) */
    for (j=0; j<4; j++) {
        for (i=0; i<6; i++) {
            if (var[j][i] != j*6+i + 50) {
#ifdef PRINT_ERR_ON_SCREEN
                /* this error is a pntecdf internal error, if occurs */
                printf("Error: bput_varm write buffer has been altered at j=%d i=%d\n",j,i);
#endif
                errs++;
                break;
            }
        }
    }
    if (NC_NOERR != (retval = ncmpi_close(ncid))) ERR(retval);

    /* check if PnetCDF freed all internal malloc */
    MPI_Offset malloc_size, sum_size;
    int err = ncmpi_inq_malloc_size(&malloc_size);
    if (err == NC_NOERR) {
        MPI_Reduce(&malloc_size, &sum_size, 1, MPI_OFFSET, MPI_SUM, 0, MPI_COMM_WORLD);
        if (rank == 0 && sum_size > 0)
            printf("heap memory allocated by PnetCDF internally has %lld bytes yet to be freed\n",
                   sum_size);
    }

    if (rank == 0) {
        char cmd_str[80];
        sprintf(cmd_str, "*** TESTING C   %s for bput API ", argv[0]);

        if (errs)
            printf("%-66s ------ failed\n", cmd_str);
        else
            printf("%-66s ------ pass\n", cmd_str);
    }
    MPI_Finalize();

    return errs;
}

